# coding: utf-8
""" 
Misc copy utilities
"""
from __future__ import unicode_literals

import filecmp
import fnmatch
import gettext
import logging
import os
import os.path
import re
import shutil
import sys
from builtins import str

from future.utils import text_to_native_str

from .exceptions import CopyException
from .exceptions import NgoFileException
from .exceptions import NotADirectoryException
from .exceptions import NotExistingPathException

enc = sys.stdout.encoding or "cp850"
_ = gettext.gettext


def _copy(src, dst):
    """
    Copy a file src to dst (directory).
    
    If dst exists and is the same, nothing is done

    :param src: source file or directory
    :type src: path
    :param dst: destination file or directory
    :type dst: path
    """
    logger = logging.getLogger(__name__)
    src = text_to_native_str(src)
    dst = text_to_native_str(dst)
    if os.path.exists(dst) and os.path.isdir(dst):
        dst = os.path.join(dst, os.path.basename(src))
    if os.path.exists(dst):
        if not filecmp.cmp(src, dst):
            logger.debug('updating ' + str(dst, enc))
            shutil.copy2(src, dst)
        else:
            pass
            # logger.info(unicode(dst,enc) + ' already up to date')
    else:
        logger.debug('copy ' + str(src, enc) + ' to ' + str(dst, enc))
        shutil.copy2(src, dst)


def _copytree(src, dst, excludes=[], includes=[], recursive=True):
    """ 
    Copy a directory structure src to destination
    
    :param src: source file or directory
    :type src: path
    :param dst: destination file or directory
    :type dst: path
    :param excludes: list of patterns to exclude
    :type excludes: ngomodel.validators.List
    :param includes: list of patterns to include
    :type includes: ngomodel.validators.List
    :param recursive: recursive copy
    """
    logger = logging.getLogger(__name__)
    # make sure to convert string from ngopath
    # convert everything to ngopath and back to string
    src = text_to_native_str(src)
    dst = text_to_native_str(dst)

    incl = r'|'.join([fnmatch.translate(x) for x in includes])
    excl = r'|'.join([fnmatch.translate(x) for x in excludes]) or r'$.'

    allnames = os.listdir(src)
    names = [name for name in allnames if not re.match(excl, name)]

    errors = []

    if not os.path.exists(dst):
        logger.debug('making dir' + str(dst, enc))
        os.makedirs(dst)

    for name in names:
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        try:
            if os.path.isdir(srcname):
                if recursive:
                    _copytree(srcname, dstname, excludes, includes, recursive)
                if re.match(incl, name) and recursive:
                    _copytree(srcname, dstname, excludes, includes, recursive)
            elif re.match(incl, name):
                _copy(srcname, dstname)
        except (IOError, os.error) as why:
            errors.append((srcname, dstname, str(why)))
    try:
        shutil.copystat(src, dst)
    except WindowsError:
        # can't copy file access times on Windows
        pass
    except OSError as why:
        errors.extend((src, dst, str(why)))
    if errors:
        raise CopyException(errors)


def advanced_copy(src,
                  dst,
                  excludes=[],
                  includes=[],
                  recursive=True,
                  create_directory=True):
    """
    Copy a directory structure src to destination
    
    :param src: source file or directory
    :type src: path
    :param dst: destination file or directory
    :type dst: path
    :param excludes: list of patterns to exclude
    :type excludes: list
    :param includes: list of patterns to include
    :type includes: list
    :param recursive: recursive copy
    :param create_directory: create missing directories
    """
    logger = logging.getLogger(__name__)

    src = text_to_native_str(src)  # not Path because it could have a pattern
    dsts = validators.TypedSet(validators.Path)(
        dst
    )  # not is dir because it might not exist if it s just being created
    includes = validators.TypedSet(text_to_native_str)(includes)
    excludes = validators.TypedSet(text_to_native_str)(excludes)

    # treat case src is given as a pattern and does not really exist,
    # convert it to an include
    if '*' in src:
        src = src.replace('\\', '/')
        bf, af = src.split('*', 1)
        src, inc = bf.rsplit('/', 1)
        inc = '%s*%s' % (inc, af)
        includes.add(inc)
    src = validators.ExistingPath(src)

    for dst in dsts:
        # do we need to create dst directories ?
        parts = dst.parts
        cur = validators.Path(parts[0])
        assert cur.exists(), '%s does not exist' % str(cur)
        for p in parts[1:]:
            cur = cur.joinpath(p)
            if not cur.is_dir():
                if not create_directory:
                    raise NotADirectoryException(
                        _('Use create_directory option.'), cur)
                logger.debug(_('creating directory %s' % cur))
                os.makedirs(text_to_native_str(cur.resolve()))
        if src.is_file():
            logger.debug('_copy(%s,%s)' % (str(src), str(dst)))
            _copy(src, dst)
        else:
            logger.debug('_copytree(%s,%s,...)' % (str(src), str(dst)))
            _copytree(src, dst, excludes, includes, recursive)



# -*- coding: utf-8 -*-
"""
Unit tests for ngofile

setup.py - created on 2017/11/15 08:44:32
author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3 """
from __future__ import unicode_literals

import logging
from builtins import object
from builtins import str
from pathlib import Path

from ngofile.pathlist import PathList

test_file = Path(__file__).resolve()
test_dir = Path(__file__).resolve().parent


class TestPathList(object):
    logger = logging.getLogger(__name__)

    def test_singleton(self):
        a = PathList(test_dir)
        b = PathList()
        b.add(str(
            test_dir))  # test_dir is already added during initialisation of a
        assert len(a.pathlist) == 1

    def test_exists(self):
        a = PathList(test_dir)
        p2 = a.pick_first(test_file.name)
        assert p2.exists()
        assert not a.exists('dummy.dum')


if __name__ == '__main__':
    TestPathList().test_singleton()
    TestPathList().test_exists()



# -*- coding: utf-8 -*-
""" 
utilities to list files in a directory, or a zip file

author: Cedric ROMAN (roman@numengo.com)
licence: GNU GPLv3
"""
from __future__ import unicode_literals

import fnmatch
import gettext
import logging
import os
import os.path
import pathlib
import re
import zipfile
from builtins import range
from builtins import str

from future.utils import text_to_native_str

from .exceptions import NotADirectoryException
from .exceptions import NotAZipArchiveException
from .exceptions import NotExistingPathException

_ = gettext.gettext


def list_files(src,
               includes=["*"],
               excludes=[],
               recursive=False,
               in_parents=False,
               folders=1,
               raise_src_exists=True):
    """
    List files in a source path with a list of given patterns

    if src contains patterns, modifies initial source dir and create corresponding includes patterns
    
    :param src: source directory
    :type src: str
    :param includes: pattern or list of patterns (*.py, *.txt, etc...)
    :type includes: [str,list]
    :param excludes: pattern or patterns to exclude
    :type excludes: [str,list]
    :param recursive: list files recursively
    :param in_parents: list files recursively in parents
    :param folders: 0: without folders, 1: with folders, 2: only folders
    :type folders: enum:[0,1,2]
    :param raise_src_exists: raise exception if src does not exist, or return empty list
    :rtype: path
    """
    logger = logging.getLogger(__name__)
    if type(src) in [list, set, tuple]:
        for s in src:
            for f in list_files(s, includes, excludes, recursive, in_parents,
                                folders, raise_src_exists):
                yield f
        return

    # declare includes regex and create a function to compile it
    global inclp
    inclp = None

    def _compile_includes(includes):
        global inclp
        incl = r'|'.join([fnmatch.translate(x.lower()) for x in includes])
        inclp = re.compile(incl)

    # declare excludes regex and create a function to compile it
    global exclp
    exclp = None

    def _compile_excludes(excludes):
        global exclp
        excl = r'|'.join([fnmatch.translate(x.lower())
                          for x in excludes]) or r'$.'
        exclp = re.compile(excl)

    # global counters
    global f_count
    global d_count
    f_count = 0
    d_count = 0

    # first we define a helper function for recursive operations
    def list_files_in_dir(srcdir, includes, excludes, recursive):
        """ we use strings and only create a path object if it s a resut """
        global inclp, exclp
        global d_count
        lf_count = 0  # local file counter
        ld_count = 0  # local inner directory counter
        for p in includes:
            p2 = p.replace('\\', '/')
            if '/' in p2:
                includes2 = set(includes)
                pa, pb = p2.split('/', 1)
                includes2.remove(p)
                includes2.add(pb)
                pa = os.path.join(srcdir, pa)
                if os.path.exists(pa):
                    ld_count += 1
                    _compile_includes(includes2)
                    for f in list_files_in_dir(pa, includes2, excludes,
                                               recursive):
                        lf_count += 1
                        yield f
                    _compile_includes(includes)

        for p in excludes:
            p2 = p.replace('\\', '/')
            if '/' in p2:
                excludes2 = set(excludes)
                pa, pb = p2.split('/', 1)
                excludes2.remove(p)
                excludes2.add(pb)
                pa = os.path.join(srcdir, pa)
                if os.path.exists(pa):
                    ld_count += 1
                    _compile_excludes(excludes2)
                    for f in list_files_in_dir(pa, includes, excludes2,
                                               recursive):
                        lf_count += 1
                        yield f
                    _compile_excludes(excludes)
        names_all = os.listdir(srcdir)
        names_not_excl = [
            name for name in names_all if not exclp.match(name.lower())
        ]
        for name in names_not_excl:
            path = os.path.join(srcdir, name)
            is_dir = os.path.isdir(path)
            if os.path.isdir(path) and recursive:
                for f in list_files_in_dir(path, includes, excludes,
                                           recursive):
                    lf_count += 1
                    yield pathlib.Path(f)
            if inclp.match(name.lower()):
                if ((folders == 0 and not is_dir) or (folders == 1)
                        or (folders == 2 and is_dir)):
                    lf_count += 1
                    yield pathlib.Path(path)

        d_count += ld_count
        logger.debug(
            _('found %i files in %s (and %i inner directories)' %
              (lf_count, srcdir, ld_count)))

    src = text_to_native_str(str(src))
    # treat case src is given as a pattern and does not really exist,
    # convert it to an include
    if '*' in src:
        src = src.replace('\\', '/')
        bf, af = src.split('*', 1)
        src, inc = bf.rsplit('/', 1)
        inc = '%s*%s' % (inc, af)
        includes.add(inc)
    if not os.path.exists(src):
        if raise_src_exists:
            raise IOError(
                _('impossible to list file in non existing directory %s' %
                  src))
        return
    if not os.path.isdir(src):  # it s a file, returns it
        yield pathlib.Path(src)
        return

    # to compile regexp only when they change
    if not isinstance(includes, list):
        includes = [includes]
    if not isinstance(excludes, list):
        excludes = [excludes]
    # this creates a set of all includes/excludes in native string
    includes = set([text_to_native_str(i) for i in includes])
    excludes = set([text_to_native_str(e) for e in excludes])
    _compile_includes(includes)
    _compile_excludes(excludes)

    for f in list_files_in_dir(src, includes, excludes, recursive):
        f_count += 1
        yield pathlib.Path(f)

    logger.info(
        _('found %i files in %s (and %i inner directories)' % (f_count, src,
                                                               d_count)))

    if in_parents:
        srcdir = pathlib.Path(src)
        cur = srcdir.resolve()
        while cur.stem:
            f_count = 0
            d_count = 0
            excludes2 = excludes.union(set([cur.relative_to(cur.parent)]))
            excludes2 = set([text_to_native_str(str(e)) for e in excludes2])
            for f in list_files_in_dir(
                    text_to_native_str(str(cur.parent)), includes, excludes2,
                    recursive):
                f_count += 1
                yield pathlib.Path(f)
            if f_count:
                logger.info(
                    _('found %i files in parents in %s (and %i inner directories)'
                      % (f_count, cur.parent, d_count)))
            cur = cur.parent


def list_files_in_zip(archive,
                      includes=["*"],
                      excludes=[],
                      recursive=False,
                      directories=True):
    """
    List files in a zip file

    :param archive: zipfile to explore
    :type archive: zipfile.ZipFile
    :param includes: pattern or list of patterns ('*.py', '*.txt', etc...)
    :type includes: [str,list]
    :param excludes: patterns to exclude
    :type excludes: [str,list]
    :param recursive: list files recursively
    :param directories: list also directories (NOT IMPLEMENTED)
    :rtype: list
    """
    if not isinstance(includes, list):
        includes = [includes]
    includes = [i.replace('\\', '/') for i in includes]
    if not isinstance(archive, zipfile.ZipFile):
        raise NotAZipArchiveException('%r is not a valid zip file' % archive)
    dirs = set()
    for d in [
            re.match('(.*)/[^/]*$', l).group(1) for l in archive.namelist()
            if re.match('(.*)/[^/]*$', l)
    ]:
        dirs.add(d)
        for i in range(len(d.split('/'))):
            dirs.add('/'.join(d.split('/')[0:i]))
    incl = r'|'.join([fnmatch.translate(x) for x in includes])
    excl = r'|'.join([fnmatch.translate(x) for x in excludes]) or r'$.'
    if len([x for x in dirs if re.match(incl, x)]):
        incl = incl.replace(
            '\Z(?ms)', r'\/.*\Z(?ms)'
        )  #http://stackoverflow.com/questions/11998613/regular-expression-zms
        incl = incl.replace('$', r'\/.*$')
    if not recursive:
        incl = incl.replace('.*\Z(?ms)', r'[^/]*\Z(?ms)')
        incl = incl.replace('.*$', r'[^/]*$')

    for n in archive.namelist():
        if re.match(incl, n) and re.search(excl, n) is None:
            yield n



# -*- coding: utf-8 -*-
""" Unit tests for _list_files

setup.py - created on 2017/11/15 08:44:32
author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3 """
from __future__ import unicode_literals

import zipfile
from builtins import str
from pathlib import Path

from ngofile.list_files import list_files
from ngofile.list_files import list_files_in_zip

test_file = Path(__file__).resolve()
test_dir = Path(__file__).resolve().parent
test_dir_a = test_dir.joinpath('a')


def test_list_files_with_patterns():
    # works with Path object
    fs1 = list(list_files(test_dir_a, recursive=True))
    # works with path as a string
    fs2 = list(list_files(str(test_dir_a), recursive=True))
    assert len(fs1) == len(fs2)
    # only selects *.py files
    fs = list(list_files(test_dir_a, "*.data", recursive=True))
    assert len(fs) == 9
    # only selects *.py and *.txt files
    fs = list(
        list_files(test_dir_a, ["*.data", "*.txt"], recursive=True))
    assert len(fs) == 18
    # only selects *.py files, but excludes b and bb directories
    fs = list(
        list_files(
            test_dir_a, "*.data", ["bb", "bbb"], recursive=True))
    assert len(fs) == 4

    assert str(test_file) == str(next(list_files(test_file)))


def test_list_files_in_zip():
    f = test_dir.joinpath('tmp_dir_py.zip')
    z = zipfile.ZipFile(str(f), 'r')
    ls1 = list_files_in_zip(z, "*.py", excludes=[], recursive=True)
    ls2 = list_files_in_zip(z, "*", excludes=["*.txt"], recursive=True)
    assert len(list(ls2)) >= len(list(ls1))  # because ls1 contains folders too
    z.close()


if __name__ == '__main__':
    test_list_files_with_patterns()



# coding: utf-8
"""
functions/class to deal with filepaths
"""
from __future__ import unicode_literals

import logging
import operator
import os.path
import pathlib
import pprint
import importlib
from builtins import str
from builtins import object

from future.utils import text_to_native_str

from .list_files import list_files


class PathList(object):
    """
    Pathlist manager
    """
    logger = logging.getLogger(__name__)

    def __init__(self, *args, **kwargs):
        """
        Appends each arg as a path of pathlist
        """
        self._pathdict = {}
        for arg in args:
            self.add(arg)

    def __str__(self):
        return ("<%s>\n%s" % (self.__class__.__name__,
                              pprint.pformat(list(self.pathlist))))

    def __repr__(self):
        '''
        Return representation of the object
        '''
        return (
            "<%s[%i items]>" % (self.__class__.__name__, len(self._pathdict)))

    @property
    def pathlist(self):
        return set(self._pathdict.keys())

    @pathlist.setter
    def pathlist(self, pathlist):
        """
        Set the pathlist from a list

        :type pathlist: list
        """
        self._pathdict.clear()
        for p in pathlist:
            self.add(p)

    def as_strings(self):
        """
        Return the pathlist as a set of string

        :rtype: list, items:{type:str}
        """
        return [text_to_native_str(str(p)) for p in self.pathlist]

    def add(self, path):
        """
        Append a path to pathlist (only if it exists)

        :param path: path to assert
        :type path: path
        """
        if type(path) in [list, set, tuple]:
            return [self.exists(p) for p in path]
        p = text_to_native_str(str(path))
        if '*' in p:
            # use list_files to create a list of files corresponding to pattern
            # and add the list pathlist
            self.add(list(list_files(p)))
        p = pathlib.Path(path)
        if p not in self.pathlist:
            if p.exists():
                p = p.resolve()
                self._pathdict[p] = 0  # intialize counter
            else:
                self.logger.warning('%s does not exist' % p)

    def add_module_path(self, module, *args):
        """
        Append a module path to pathlist

        :param module: module name
        :type module: string
        """
        m = importlib.import_module(module)
        self.add(pathlib.Path(m.__file__).parent.joinpath(args))

    def exists(self, path):
        """
        Check if a path exists in pathlist

        :param path: path or pattern
        :type path: path
        :rtype: bool
        """
        if type(path) in [list, set, tuple]:
            return [self.exists(p) for p in path]
        return bool(self.pick_first(path))

    def pick_first(self, path):
        """
        Pick the first existing match

        :param path: path or pattern
        :type path: str
        :rtype: path
        """
        if type(path) in [list, set, tuple]:
            return [self.pick_first(p) for p in path]
        path = text_to_native_str(str(path))
        path = path.replace('\\', '/')
        includes = []
        if '*' in path:
            bf, af = path.split('*', 1)
            path, inc = bf.rsplit('/', 1)
            includes = ['%s*%s' % (inc, af)]
        if os.path.exists(path):
            return pathlib.Path(path)
        optimized_pathlist = sorted(
            list(self._pathdict.items()),
            key=operator.itemgetter(1),
            reverse=True)
        for p, _oldc in optimized_pathlist:
            if p.joinpath(path).exists():
                return next(list_files(p.joinpath(path), includes))

    def list_files(self,
                   includes=["*"],
                   excludes=[],
                   recursive=False,
                   in_parents=False,
                   flatten=True):
        """
        List files in a source directory with a list of given patterns 
        
        :param includes: pattern or list of patterns ('*.py', '*.txt', etc...)
        :type includes: [str,list]
        :param excludes: patterns to exclude
        :type excludes: [str,list]
        :param recursive:list files recursively
        :param in_parents: list files recursively in parents
        :param flatten: flatten return lists
        :rtype: array, items:{type: path}
        """
        ret = [None] * len(self._pathdict)
        optimized_pathlist = sorted(
            list(self._pathdict.items()),
            key=operator.itemgetter(1),
            reverse=True)
        for p, _oldc in optimized_pathlist:
            lf = list_files(p, includes, excludes, recursive, in_parents)
            if flatten:
                for f in lf:
                    yield f
            else:
                yield list(lf)



# *- coding: utf-8 -*-
"""click API for NgoFile"""
import click
import gettext

from ngofile.list_files import list_files
from ngofile.copy import advanced_copy

_ = gettext.gettext


@click.command()
@click.argument('names', nargs=-1)
def main(names):
    click.echo(repr(names))


########################
# FUNCTIONS ENTRY POINTS
########################


@click.command()
@click.argument('src')
@click.option(
    '--includes',
    default=[u'*'],
    help=_('pattern or list of patterns (*.py, *.txt, etc...)'))
@click.option('--excludes', help=_('pattern or patterns to exclude'))
@click.option('--recursive', help=_('list files recursively'))
@click.option('--in_parents', help=_('list files recursively in parents'))
@click.option(
    '--folders',
    default=1,
    help=_('0: without folders, 1: with folders, 2: only folders'))
@click.option(
    '--raise_src_exists',
    default=True,
    help=_('raise exception if src does not exist, or return empty list'))
def list_files_cli(ctx, src, includes, excludes, recursive, in_parents,
                   folders, raise_src_exists):
    """ list files in a source path with a list of given patterns 
    if src contains patterns, modifies initial source dir and create corresponding includes patterns"""
    list_files()


@click.command()
@click.argument('src')
@click.argument('dst')
@click.option('--excludes', help=_('list of patterns to exclude'))
@click.option('--includes', help=_('list of patterns to include'))
@click.option('--recursive', default=True, help=_('recursive copy'))
@click.option(
    '--create_directory', default=True, help=_('create missing directories'))
def advanced_copy_cli(ctx, src, dst, excludes, includes, recursive,
                      create_directory):
    """ copy a directory structure src to destination """
    advanced_copy()



from __future__ import unicode_literals

#from ngofile.cli import main

#def test_main():
#    runner = CliRunner()
#    result = runner.invoke(main, [])
#
#    assert result.output == '()\n'
#    assert result.exit_code == 0



# -*- coding: utf-8 -*-
"""
All exceptions used in the ngofile code base are defined here.
"""
from __future__ import unicode_literals

import gettext
from builtins import str

_ = gettext.gettext


class NgoFileException(Exception):
    """
    Base exception class. All ngofile-specific exceptions should subclass
    this class.
    """


class CopyException(NgoFileException):
    """
    Raised when error occurs during copy
    """


class NotExistingPathException(NgoFileException, IOError):
    """
    Raised when a path does not exist
    """

    def __init__(self, message, path=None):
        if path is not None:
            message = (_('path %(path)s does not exist' % {'path': str(path)}),
                       message)
        super(NotExistingPathException, self).__init__(message)


class NotADirectoryException(NgoFileException, ValueError):
    """
    Raised when a path is not a directory
    """

    def __init__(self, message, path=None):
        if path is not None:
            message = (_(
                'path %(path)s is not a directory' % {'path': str(path)}),
                       message)
        super(NotADirectoryException, self).__init__(message)


class NotAZipArchiveException(NgoFileException, ValueError):
    """
    Raised when the object given is not a zipfile
    """
